--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--CONTROL4 - LUTRON RA2 SELECT SWITCH
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

if (PersistData == nil) then
    PersistData = {}
end

MAIN_BINDING        = 1
LIGHT_PROXY_BINDING = 5001
ON_BINDING          = 300
OFF_BINDING         = 301
TOGGLE_BINDING      = 302

gDimmerLevel        = 0
gRaiseDimmer        = true
gClickRampRateDown  = PersistData["ratedown"] or 2000
gClickRampRateUp    = PersistData["rateup"] or 1000
gPresetLevel        = PersistData["preset"] or 100
gRampTimer          = 0
gToggleRampUp       = false

sceneCollection     = PersistData["sceneCollection"] or {}
flashCollection     = PersistData["flashCollection"] or {}
elementCounter      = 0
currentScene        = 0
executeElementTimer = 0

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--PROPERTIES HANDLER
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function OnPropertyChanged(strProperty)
    dbg("strProperty = "  .. strProperty .. " changed to: " .. Properties[strProperty])
    local strProp = string.gsub(strProperty, " ", "")
    if (PROPS[strProp] ~= nil and type(PROPS[strProp]) == "function") then
        PROPS[strProp](Properties[strProperty])
    else
        dbg("UNKNOWN | THROW AWAY: " .. strProperty)
    end
end

PROPS = {}

function PROPS.DebugMode(value)
    if (value == "Off") then
        g_debugprint = false
        g_debugunittest = false
        g_debuglog = false
        g_DebugTimer = 0
    elseif (value == "Print") then
        g_debugprint = true
        g_debugunittest = false
        g_debuglog = false
        startDebugTimer()
    elseif (value == "Log") then
        g_debugprint = false
        g_debugunittest = true
        g_debuglog = true
        startDebugTimer()
    elseif (value == "Print and Log") then
        g_debugprint = true
        g_debugunittest = true
        g_debuglog = true
        startDebugTimer()
    end
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--TIMER CALLBACKS
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function OnTimerExpired(idTimer)
    if (idTimer == g_DebugTimer) then
        dbg("Turning Debug Mode Off (timer expired)")
        C4:UpdateProperty("Debug Mode", "Off")
    elseif (idTimer == executeElementTimer) then
        playScene()
    else
        dbg('OnTimerExpired', idTimer)
        local timer = pullTimer(idTimer)
        if (not timer) then dbg("UNKNOWN TIMER FIRED | THROWING AWAY") return end
        timer.handler()
    end
end

----

g_timers = {}

function startTimer(value, units, callback)
    local timerId = C4:AddTimer(tonumber(value), units)
    g_timers[timerId] = { handler = callback }
    return timerId
end

function pullTimer(timerId)
    local timer = g_timers[timerId]
    g_timers[timerId] = nil
    return timer
end

function killTimer(timerId)
    local timer = g_timers[timerId]
    if (timer) then
        C4:KillTimer(timerId)
        g_timers[timerId] = nil
    end
    return timer
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--COMMAND HANDLER
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function ExecuteCommand(strCommand, tParams)
    dbg("EXECUTECOMMAND ON: " .. strCommand)
    if (tParams == nil) then
        tParams = {}
    end

    if (strCommand == "LIGHT_LEVEL_CHANGED") then
        local lightLevel = tParams["LEVEL"]
        gDimmerLevel = lightLevel
        SendToProxy(LIGHT_PROXY_BINDING, "LIGHT_LEVEL", lightLevel)
    elseif (strCommand == "Raise Dimmer") then
        SendLutronCommand("#OUTPUT","2")
    elseif (strCommand == "Lower Dimmer") then
        SendLutronCommand("#OUTPUT","3")
    elseif (strCommand == "Stop Dimmer") then
        SendLutronCommand("#OUTPUT","4")
    else
        dbg("UNKNOWN | THROW AWAY: " .. strCommand)
    end
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--PROXY MESSAGE HANDLER
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function SendToProxy(idBinding, strCommand, tParams, ...)
    dbgFunction("SendToProxy (" .. idBinding .. ", " .. strCommand .. ")")
    C4:SendToProxy(idBinding, strCommand, tParams)
end

function ReceivedFromProxy(idBinding, strCommand, tParams)
    dbgFunction("ReceivedFromProxy (" .. idBinding .. ", " .. strCommand .. ")")

    if (tParams == nil) then
        tParams = {}
    end

    if (idBinding == LIGHT_PROXY_BINDING) then
        if (LIGHT_PROXY_COMMAND_HANDLER[strCommand] ~= nil and type(LIGHT_PROXY_COMMAND_HANDLER[strCommand]) == "function") then
            LIGHT_PROXY_COMMAND_HANDLER[strCommand](tParams)
        else
            dbg("UNKNOWN | THROW AWAY: MESSAGE: " .. strCommand)
        end
    elseif (idBinding == MAIN_BINDING) then
        SendToProxy(LIGHT_PROXY_BINDING, "ONLINE_CHANGED", {STATE=true})
        if (strCommand == "MESSAGE") then
            if (MESSAGE_HANDLER[tParams["command"]] ~= nil and type(MESSAGE_HANDLER[tParams["command"]]) == "function") then
                MESSAGE_HANDLER[tParams["command"]](tParams)
            else
                dbg("UNKNOWN | THROW AWAY: MESSAGE: " .. tParams["command"])
            end
        else
            dbg("UNKNOWN | THROW AWAY: " .. strCommand)
        end
    elseif (idBinding == ON_BINDING or idBinding == OFF_BINDING or idBinding == TOGGLE_BINDING) then
        if (BUTTON_LINKS[strCommand]) then
            BUTTON_LINKS[strCommand](idBinding, tParams)
        end
    else
        dbg("UNKNOWN BINDING | THROW AWAY: " .. strCommand)
    end
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--MESSAGE HANDLER (FROM MAIN REPEATER)
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

MESSAGE_HANDLER = {}

function MESSAGE_HANDLER.OUTPUT(tParams)
    if (MESSAGE_HANDLER["OUTPUT_ACTION" .. tParams["action"]] ~= nil and type(MESSAGE_HANDLER["OUTPUT_ACTION" .. tParams["action"]]) == "function") then
        MESSAGE_HANDLER["OUTPUT_ACTION" .. tParams["action"]](tParams)
    else
        dbg("UNKNOWN ACTION | THROW AWAY: " .. tParams["action"])
    end
end

function MESSAGE_HANDLER.OUTPUT_ACTION1(tParams)
    dbg("HANDLE ACTION 1: LIGHT_LEVEL_CHANGED")
    local lightLevel = tParams["param1"]
    gDimmerLevel = lightLevel
    SendToProxy(LIGHT_PROXY_BINDING, "LIGHT_LEVEL", lightLevel)
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--LIGHT PROXY COMMAND HANDLER
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

LIGHT_PROXY_COMMAND_HANDLER = {}

function LIGHT_PROXY_COMMAND_HANDLER.GET_LIGHT_LEVEL(tParams)
    GET_LIGHT_LEVEL(tParams)
end

function LIGHT_PROXY_COMMAND_HANDLER.GET_CONNECTED_STATE(tParams)
    SendToProxy(LIGHT_PROXY_BINDING, "ONLINE_CHANGED", {STATE=true})
end

function LIGHT_PROXY_COMMAND_HANDLER.ON(tParams)
  SET_LEVEL({LEVEL=100})
end

function LIGHT_PROXY_COMMAND_HANDLER.OFF(tParams)
  SET_LEVEL({LEVEL=0})
end

function LIGHT_PROXY_COMMAND_HANDLER.BUTTON_ACTION(tParams)
    local action = tParams["ACTION"]
    if (action == "2") then
        TOGGLE_PRESET(tParams)
    else
        dbg("UNKNOWN BTN ACTION | THROW AWAY: " .. tParams["ACTION"])
    end
end

function LIGHT_PROXY_COMMAND_HANDLER.SET_CLICK_RATE_DOWN(tParams)
    gClickRampRateDown = tParams["RATE"]
    dbg("CLICK RAMP RATE DOWN = " .. gClickRampRateDown)
    PersistData["ratedown"] = gClickRampRateDown
end

function LIGHT_PROXY_COMMAND_HANDLER.SET_CLICK_RATE_UP(tParams)
    gClickRampRateUp = tParams["RATE"]
    dbg("CLICK RAMP RATE UP = " .. gClickRampRateUp)
    PersistData["rateup"] = gClickRampRateUp
end

function LIGHT_PROXY_COMMAND_HANDLER.PUSH_SCENE(tParams)
    for k,v in pairs(tParams) do
        dbg(k .. ": " .. v)
    end
    local sceneNum = tParams["SCENE_ID"]
    local elements = tParams["ELEMENTS"]
    local flash = tParams["FLASH"]
    dbg("scene_id" .. sceneNum)
    dbg("elements: " .. tParams["ELEMENTS"])
    local elementTable = collect(elements)
    local scene = {}
    for i=1,#elementTable do
        local t = {}
        t["Delay"] = elementTable[i][1][1]
        t["Rate"] = elementTable[i][2][1]
        t["Level"] = elementTable[i][3][1]
        table.insert(scene,t)
    end
    sceneCollection[sceneNum] = scene
    flashCollection[sceneNum] = flash
    PersistData["sceneCollection"] = sceneCollection
    PersistData["flashCollection"] = flashCollection
end

function LIGHT_PROXY_COMMAND_HANDLER.REMOVE_SCENE(tParams)
    local sceneNum = tParams["SCENE_ID"]
    dbg("scene_id:" .. sceneNum)
    sceneCollection[sceneNum] = nil
    flashCollection[sceneNum] = nil
    PersistData["sceneCollection"] = sceneCollection
    PersistData["flashCollection"] = flashCollection
end

function LIGHT_PROXY_COMMAND_HANDLER.ACTIVATE_SCENE(tParams)
    local sceneNum = tParams["SCENE_ID"]
    for k,v in pairs(tParams) do
        dbg(k .. ": " .. v)
    end
    currentScene = sceneNum
    elementCounter = 0
    playScene()
end

function LIGHT_PROXY_COMMAND_HANDLER.SET_PRESET_LEVEL(tParams)
    gPresetLevel = tParams["LEVEL"]
    PersistData["preset"] = gPresetLevel
end

function LIGHT_PROXY_COMMAND_HANDLER.RAMP_SCENE_UP(tParams)
    --Dimmer Start Raise
    SendLutronCommand("#OUTPUT","2")
end

function LIGHT_PROXY_COMMAND_HANDLER.RAMP_SCENE_DOWN(tParams)
    --Dimmer Start Lower
    SendLutronCommand("#OUTPUT","3")
end

function LIGHT_PROXY_COMMAND_HANDLER.STOP_SCENE_RAMP(tParams)
    --Dimmer Stop
    SendLutronCommand("#OUTPUT","4")
end

function LIGHT_PROXY_COMMAND_HANDLER.RAMP_TO_LEVEL(tParams)
    SET_LEVEL(tParams)
end

function LIGHT_PROXY_COMMAND_HANDLER.SET_LEVEL(tParams)
    SET_LEVEL(tParams)
end

function LIGHT_PROXY_COMMAND_HANDLER.TOGGLE(tParams)
    TOGGLE_PRESET(tParams)
end

BUTTON_LINKS = {}

function BUTTON_LINKS.DO_PUSH(id, tParams)
    if (id == TOGGLE_BINDING) then
        gToggleRampUp = not gToggleRampUp
    end

    local cmd = "3"
    if (gToggleRampUp) then
        cmd = "2"
    end

    gRampTimer = startTimer(
        750,
        'MILLISECONDS',
        function()
            SendLutronCommand("#OUTPUT", cmd)
        end)
end

function BUTTON_LINKS.DO_RELEASE(id, tParams)
    gRampTimer = killTimer(gRampTimer)

    LIGHT_PROXY_COMMAND_HANDLER.STOP_SCENE_RAMP()
end

function BUTTON_LINKS.DO_CLICK(id, tParams)
    gRampTimer = killTimer(gRampTimer)

    if (id == ON_BINDING) then
        SET_LEVEL({LEVEL=100})
    elseif (id == OFF_BINDING) then
        SET_LEVEL({LEVEL=0})
    elseif (id == TOGGLE_BINDING) then
        TOGGLE_PRESET(tParams)
    end
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--ADV LIGHTING SCENES
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function playScene()
    dbg("playScene")
    if (elementCounter ~= 0) then
        local t = {}
        t["LEVEL"] = sceneCollection[tostring(currentScene)][elementCounter]["Level"]
        t["TIME"] = sceneCollection[tostring(currentScene)][elementCounter]["Rate"]
        SET_LEVEL(t)
    end
    elementCounter = elementCounter + 1
    dbg(elementCounter)

    if (elementCounter > #sceneCollection[tostring(currentScene)] and flashCollection[tostring(currentScene)] == "1") then
        elementCounter = 1
    end

    if (elementCounter <= #sceneCollection[tostring(currentScene)]) then
        local timeInterval = sceneCollection[tostring(currentScene)][elementCounter]["Delay"] or -1
        executeElementTimer = C4:KillTimer(executeElementTimer)
        executeElementTimer = C4:AddTimer(timeInterval, "MILLISECONDS")
    else
        dbg("end of scene")
    end

end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--SEND LUTRON COMMAND
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function SendLutronCommand(strCommand, action, param1, param2)
    dbg("EXECUTE LUTRON CMD: " .. strCommand .. " ACTION: " .. action)
    local t = { command = strCommand, action = action, param1 = param1, param2 = param2 }
    C4:SendToProxy(MAIN_BINDING, "LUTRON_CMD", t)
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--LIGHT LEVEL HANDLING
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function GET_LIGHT_LEVEL(tParams)
    lightLevel = gDimmerLevel
    C4:SendToProxy(LIGHT_PROXY_BINDING, "LIGHT_LEVEL_CHANGED", lightLevel)
    SendLutronCommand("?OUTPUT","1")
end

function localizedNumber(input)
    if (type(input) == "number") then
        return input
    elseif (type(input) == "string") then
        input = string.gsub(input,",",".")
        local output = tonumber(input)
        return output
    end
end

function SET_LEVEL(tParams)
    if (tParams ~= nil) then
        if (tParams["LEVEL"] ~= nil) then
            local param2 = ""
            local rate = 0
            if (tParams["TIME"] ~= nil) then
                local tempValue = (tonumber(tParams["TIME"]))/1000   -- convert to seconds
                rate = localizedNumber(tempValue)
            end

            if (rate ~= 0) then
                param2 = string.format("%02d:%02d:%02d", rate/3600, (rate/60)%60, rate%60)
            else
                param2 = "0.25"
            end
            SendLutronCommand("#OUTPUT","1", tostring(tParams["LEVEL"]), param2)
        end
    end
end

function TOGGLE_PRESET(tParams)
    if (tonumber(gDimmerLevel) == 0) then  -- this ignores presets and should be changed if presets are implemented
        tParams["LEVEL"] = gPresetLevel
        tParams["TIME"] = gClickRampRateUp
    else
        tParams["LEVEL"] = 0
        tParams["TIME"] = gClickRampRateDown
    end
    SET_LEVEL(tParams)
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--DEBUG
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function startDebugTimer()
    if (g_DebugTimer) then
        g_DebugTimer = C4:KillTimer(g_DebugTimer)
    end
    g_DebugTimer = C4:AddTimer(45, "MINUTES")
end

function dbg(strDebugText)
    if (g_debugprint) then print(strDebugText) end
    if (g_debuglog) then C4:ErrorLog("Lutron_RA2_Select_Switch: " .. strDebugText) end
end

function dbgUnitTest(strDebugText)
    if (g_debugunittest) then print (strDebugText) end
    if (g_debuglog) then C4:ErrorLog("Lutron_RA2_Select_Switch: " .. strDebugText) end
end

function dbgFunction(strFunction)
    local debugFunction = true
    if (g_debugprint and debugFunction) then print("Function:\t" .. strFunction .. ":") end
    if (g_debuglog  and debugFunction) then C4:ErrorLog("Lutron_RA2_Select_Switch: " .. strFunction) end
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--XML PARSER (ADV LIGHTING SCENES)
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

function parseargs(s)
  local arg = {}
  string.gsub(s, "(%w+)=([\"'])(.-)%2", function (w, _, a)
    arg[w] = a
  end)
  return arg
end

function collect(s)
  local stack = {}
  local top = {}
  table.insert(stack, top)
  local ni,c,label,xarg, empty
  local i, j = 1, 1
  while true do
    ni,j,c,label,xarg, empty = string.find(s, "<(%/?)([%w:]+)(.-)(%/?)>", i)
    if not ni then break end
    local text = string.sub(s, i, ni-1)
    if not string.find(text, "^%s*$") then
      table.insert(top, text)
    end
    if empty == "/" then  -- empty element tag
      table.insert(top, {label=label, xarg=parseargs(xarg), empty=1})
    elseif c == "" then   -- start tag
      top = {label=label, xarg=parseargs(xarg)}
      table.insert(stack, top)   -- new level
    else  -- end tag
      local toclose = table.remove(stack)  -- remove top
      top = stack[#stack]
      if #stack < 1 then
        error("nothing to close with "..label)
      end
      if toclose.label ~= label then
        error("trying to close "..toclose.label.." with "..label)
      end
      table.insert(top, toclose)
    end
    i = j+1
  end
  local text = string.sub(s, i)
  if not string.find(text, "^%s*$") then
    table.insert(stack[#stack], text)
  end
  if #stack > 1 then
    error("unclosed "..stack[#stack].label)
  end
  return stack[1]
end

--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--CLEANUP
--//////////--//////////--//////////--//////////--//////////--//////////--//////////
--//////////--//////////--//////////--//////////--//////////--//////////--//////////

-- Release things this driver had allocated...
function OnDriverDestroyed()
    -- Kill open timers
    if (g_DebugTimer ~= nil) then g_DebugTimer = C4:KillTimer(g_DebugTimer) end
end

function OnVariableChanged(strName)
    dbgFunction("OnVariableChanged(" .. strName .. ")")
end
